@@ -68,6 +68,7 @@ gksu_error_get_type (void)
       { GKSU_ERROR_WRONGPASS, "GKSU_ERROR_WRONGPASS", "wrongpass" },
       { GKSU_ERROR_CHILDFAILED, "GKSU_ERROR_CHILDFAILED", "childfailed" },
       { GKSU_ERROR_CANCELED, "GKSU_ERROR_CANCELED", "canceled" },
+      { GKSU_ERROR_WRONGAUTOPASS, "GKSU_ERROR_WRONGAUTOPASS", "wrongautopass" },
       { 0, NULL, NULL }
     };
     etype = g_enum_register_static ("GksuError", values);
@@ -657,8 +658,6 @@ get_gnome_keyring_password (GksuContext *context)
       password = g_locale_from_utf8 (found->secret,
 				     password_length,
 				     NULL, NULL, NULL);
-      /* strip the \n that is always stored in gnome-keyring */
-      password[password_length - 1] = '\0';
       return password;
     }
 
@@ -904,6 +903,11 @@ su_ask_password (GksuContext *context, gchar *prompt,
 		       GKSU_ERROR_CANCELED,
 		       _("Password prompt canceled."));
 	}
+
+      gtk_widget_destroy (dialog);
+      while (gtk_events_pending ())
+	gtk_main_iteration ();
+
       return NULL;
     }
 
@@ -911,6 +915,10 @@ su_ask_password (GksuContext *context, gchar *prompt,
   password = g_locale_from_utf8 (tmp, strlen (tmp), NULL, NULL, NULL);
   g_free (tmp);
 
+  gtk_widget_destroy (dialog);
+  while (gtk_events_pending ())
+    gtk_main_iteration ();
+
   return password;
 }
 
@@ -1048,26 +1056,13 @@ get_process_name (pid_t pid)
   return strdup(buf.cmd);
 }
 
-/**
- * prepare_xauth:
- *
- * Sets up the variables with values for the $DISPLAY
- * environment variable and xauth-related stuff. Also
- * creates a temporary directory to hold a .Xauthority
- *
- * Returns: TRUE if it suceeds, FALSE if it fails.
- */
-static int
-prepare_xauth (GksuContext *context)
+static gchar *
+get_xauth_token (GksuContext *context, gchar *display)
 {
   gchar *xauth_bin = NULL;
-
   FILE *xauth_output;
-  gchar xauth[256] = {0};
-
   gchar *tmp = NULL;
-
-  context->display = g_strdup (getenv ("DISPLAY"));
+  gchar *xauth = g_new0 (gchar, 256);
 
   /* find out where the xauth binary is located */
   if (g_file_test ("/usr/bin/xauth", G_FILE_TEST_IS_EXECUTABLE))
@@ -1080,32 +1075,83 @@ prepare_xauth (GksuContext *context)
 	       "Failed to obtain xauth key: xauth binary not found "
 	       "at usual locations");
 
-      return FALSE;
+      return NULL;
     }
 
   /* get the authorization token */
   tmp = g_strdup_printf ("%s list %s | "
 			 "head -1 | awk '{ print $3 }'",
 			 xauth_bin,
-			 context->display);
+			 display);
   if ((xauth_output = popen (tmp, "r")) == NULL)
     {
       fprintf (stderr,
 	       "Failed to obtain xauth key: %s",
 	       strerror(errno));
-      return FALSE;
+      return NULL;
     }
   fread (xauth, sizeof(char), 255, xauth_output);
   pclose (xauth_output);
   g_free (tmp);
 
-  context->xauth = g_strdup (xauth);
-
   if (context->debug)
     {
       fprintf(stderr,
 	      "xauth: -%s-\n"
 	      "display: -%s-\n",
+	      xauth, display);
+    }
+
+  return xauth;
+}
+
+/**
+ * prepare_xauth:
+ *
+ * Sets up the variables with values for the $DISPLAY
+ * environment variable and xauth-related stuff. Also
+ * creates a temporary directory to hold a .Xauthority
+ *
+ * Returns: TRUE if it suceeds, FALSE if it fails.
+ */
+static int
+prepare_xauth (GksuContext *context)
+{
+  gchar *display = NULL;
+  gchar *xauth = NULL;
+
+  display = g_strdup (getenv ("DISPLAY"));
+  xauth = get_xauth_token (context, display);
+  if (xauth == NULL)
+    {
+      g_free (display);
+      return FALSE;
+    }
+
+  /* If xauth is the empty string, then try striping the
+   * hostname part of the DISPLAY string for getting the
+   * auth token; this is needed for ssh-forwarded usage
+   */
+  if (!strcmp ("", xauth))
+    {
+      gchar *cut_display = NULL;
+
+      g_free (xauth);
+      cut_display = g_strdup (g_strrstr (display, ":"));
+      xauth = get_xauth_token (context, cut_display);
+
+      g_free (display);
+      display = cut_display;
+    }
+
+  context->xauth = xauth;
+  context->display = display;
+
+  if (context->debug)
+    {
+      fprintf(stderr,
+	      "final xauth: -%s-\n"
+	      "final display: -%s-\n",
 	      context->xauth, context->display);
     }
 
@@ -1833,38 +1879,42 @@ gksu_su_full (GksuContext *context,
       /* no need to ask for password if we're already root */
       if (my_uid != target_uid && my_uid)
 	{
+	  gint count;
 	  struct termios tio;
 
-	  usleep (100);
 	  read (fdpty, buf, 255);
 	  if (context->debug)
 	    fprintf (stderr, "gksu_context_run: buf: -%s-\n", buf);
 
-	  usleep (100);
-
+	  /* make sure we notice that ECHO is turned off, if it gets
+	     turned off */
 	  tcgetattr (fdpty, &tio);
+	  for (count = 0; (tio.c_lflag & ECHO) && count < 15; count++)
+	    {
+	      usleep (30);
+	      tcgetattr (fdpty, &tio);
+	    }
 
 	  if (!(tio.c_lflag & ECHO))
 	    {
-	      gchar *tmp = NULL;
-
 	      gboolean prompt_grab;
 	      prompt_grab = gconf_client_get_bool (context->gconf_client, BASE_PATH "prompt",
 						   NULL);
+
 	      if (prompt_grab)
 		gksu_prompt_grab (context);
 
 	      /* try to get the password from the GNOME Keyring first */
-	      tmp = get_gnome_keyring_password (context);
+	      password = get_gnome_keyring_password (context);
 	      if (context->debug)
-		fprintf (stderr, "tmp: %s\n", tmp);
-	      if (tmp == NULL)
-		tmp = ask_pass (context, buf, ask_pass_data, error);
+		fprintf (stderr, "password: %s\n", password);
+	      if (password == NULL)
+		password = ask_pass (context, buf, ask_pass_data, error);
 	      else
 		used_gnome_keyring = TRUE;
 	      if (context->debug)
-		fprintf (stderr, "tmp: %s\n", tmp);
-	      if (tmp == NULL || (error && (*error)))
+		fprintf (stderr, "password: %s\n", password);
+	      if (password == NULL || (error && (*error)))
 		{
 		  if (context->debug)
 		    fprintf (stderr, "gksu_su_full: problem getting password - getting out\n");
@@ -1874,12 +1924,8 @@ gksu_su_full (GksuContext *context,
 		  return TRUE;
 		}
 
-	      usleep (100);
-
-	      password = g_strdup_printf ("%s\n", tmp);
-	      nullify_password (tmp);
-
-	      write (fdpty, password, strlen(password));
+	      write (fdpty, password, strlen(password) + 1);
+	      write (fdpty, "\n", 1);
 	      password_needed = TRUE;
 	    }
 	}
@@ -1920,9 +1966,16 @@ gksu_su_full (GksuContext *context,
 
 	      strings = g_strsplit (buf, ":", 2);
 	      if (strings[1] && !strncmp (strings[1], " Authentication failure", 23))
-		g_set_error (error, gksu_quark,
-			     GKSU_ERROR_WRONGPASS,
-			     _("Wrong password."));
+		{
+		  if (used_gnome_keyring)
+		    g_set_error (error, gksu_quark,
+				 GKSU_ERROR_WRONGAUTOPASS,
+				 _("Wrong password got from keyring."));
+		  else
+		    g_set_error (error, gksu_quark,
+				 GKSU_ERROR_WRONGPASS,
+				 _("Wrong password."));
+		}
 	      g_strfreev (strings);
 
 	      if (context->debug)
@@ -2142,6 +2195,22 @@ gksu_su (gchar *command_line, GError **error)
   return retval;
 }
 
+static void
+read_line (int fd, gchar *buffer, int n)
+{
+  gint counter = 0;
+  gchar tmp[2] = {0};
+
+  for (; counter < (n - 1); counter++)
+    {
+      read (fd, tmp, 1);
+      if (tmp[0] == '\n')
+	break;
+      buffer[counter] = tmp[0];
+    }
+  buffer[counter+1] = '\0';
+}
+
 /**
  * gksu_sudo_full:
  * @context: a #GksuContext
@@ -2189,7 +2258,6 @@ gksu_sudo_full (GksuContext *context,
 
   pid_t pid;
   int status;
-  size_t r;
   FILE *infile, *outfile;
   int parent_pipe[2];	/* For talking to the parent */
   int child_pipe[2];	/* For talking to the child */
@@ -2384,6 +2452,7 @@ gksu_sudo_full (GksuContext *context,
     }
   else
     {
+      gint counter = 0;
       gchar *cmdline = NULL;
 
       // Parent
@@ -2430,10 +2499,23 @@ gksu_sudo_full (GksuContext *context,
 	select (parent_pipe[0] + 1, &rfds, NULL, NULL, &tv);
       }
 
-      r = read(parent_pipe[0], buffer, 16);
+      /* Try hard to find the prompt; it may happen that we're
+       * seeing sudo's lecture, or that some pam module is spitting
+       * debugging stuff at the screen
+       */
+      for (counter = 0; counter < 50; counter++)
+	{
+	  if (strncmp (buffer, "GNOME_SUDO_PASS", 15) == 0)
+	    break;
+
+	  read_line (parent_pipe[0], buffer, 256);
+
+	  if (context->debug)
+	    fprintf (stderr, "buffer: -%s-\n", buffer);
+	}
 
       if (context->debug)
-	fprintf (stderr, "buffer: -%s-\n", buffer);
+	fprintf (stderr, "brute force GNOME_SUDO_PASS ended...\n");
 
       if (strncmp(buffer, "GNOME_SUDO_PASS", 15) == 0)
 	{
